# Generated by Django 4.2.23 on 2025-08-08 08:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pretix_wallet', '0002_alter_wallet_pan_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Wallet_SettingsStore',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('key', models.CharField(max_length=255)),
                ('value', models.TextField()),
                ('object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_settings_objects', to='pretix_wallet.wallet')),
            ],
        ),
    ]



# Generated by Django 4.2.23 on 2025-07-29 17:30

from django.db import migrations, models
import django.db.models.deletion
import pretix.base.models.base


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('pretixbase', '0184_customer'),
    ]

    operations = [
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('pan', models.CharField(max_length=19)),
                ('currency', models.CharField(max_length=10)),
                ('customer', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wallet', to='pretixbase.customer')),
                ('issuer', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='wallets', to='pretixbase.organizer')),
                ('order_position', models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wallet', to='pretixbase.orderposition')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, pretix.base.models.base.LoggingMixin),
        ),
        migrations.CreateModel(
            name='WalletTransaction',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('value', models.DecimalField(decimal_places=2, max_digits=13)),
                ('descriptor', models.TextField(default='')),
                ('data', models.JSONField(default=dict)),
                ('order_payment', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wallet_transactions', to='pretixbase.orderpayment')),
                ('order_position', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wallet_transactions', to='pretixbase.orderposition')),
                ('order_refund', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wallet_transactions', to='pretixbase.orderrefund')),
                ('wallet', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='transactions', to='pretix_wallet.wallet')),
            ],
        ),
        migrations.CreateModel(
            name='WalletItem',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('issue_wallet_balance', models.BooleanField(default=False)),
                ('item', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='wallet', to='pretixbase.item')),
            ],
        ),
    ]



import typing
import asn1tools
import pathlib
from pretix_uic_barcode.elements import UICBarcodeElement, VASElement, BaseBarcodeElementGenerator, BaseVASElementGenerator
from pretix.base.templatetags.money import money_filter
from pretix.base.models import OrderPosition, Order
from pretix.multidomain.urlreverse import build_absolute_uri
from . import models

ROOT = pathlib.Path(__file__).parent
BARCODE_CONTENT = asn1tools.compile_files([ROOT / "asn1" / "pretixWallet.asn"], codec="uper")

class WalletBarcodeElement(UICBarcodeElement):
    def __init__(self, wallet_id: str, issuer: str):
        self.wallet_id = wallet_id
        self.issuer = issuer

    @staticmethod
    def tlb_record_id():
        return "5101PW"

    @staticmethod
    def dosipas_record_id():
        return "_5101PXW"

    def record_content(self) -> bytes:
        return BARCODE_CONTENT.encode("PretixWallet", {
            "pan": self.wallet_id,
            "issuer": self.issuer,
        })

class WalletVASElement(VASElement):
    def __init__(self, wallet_id: str):
        self.wallet_id = wallet_id

    @staticmethod
    def record_id():
        return "W"

    def record_content(self) -> bytes:
        return BARCODE_CONTENT.encode("VASPretixWallet", {
            "pan": self.wallet_id,
        })

def find_wallet(order_position: OrderPosition, order: Order) -> typing.Optional[models.Wallet]:
    if hasattr(order_position, "wallet"):
        return order_position.wallet
    elif hasattr(order.customer, "wallet"):
        return order.customer.wallet
    else:
        qs = models.Wallet.objects.filter(order_position__order=order).distinct()
        if qs.count() == 1:
            return qs.first()

    return None

class WalletBarcodeElementGenerator(BaseBarcodeElementGenerator):
    @staticmethod
    def generate_element(order_position: OrderPosition, order: Order) -> typing.Optional[WalletBarcodeElement]:
        wallet = find_wallet(order_position, order)
        if not wallet:
            return None

        return WalletBarcodeElement(
            wallet_id=wallet.pan,
            issuer=wallet.issuer.slug
        )

class WalletVASElementGenerator(BaseVASElementGenerator):
    @staticmethod
    def generate_vas_element(order_position: OrderPosition, order: Order) -> typing.Optional[WalletVASElement]:
        wallet = find_wallet(order_position, order)
        if not wallet:
            return None

        return WalletVASElement(
            wallet_id=wallet.pan,
        )


def generate_google_wallet_module(order_position: OrderPosition, order: Order):
    wallet = find_wallet(order_position, order)
    if not wallet:
        return []

    return [("valueAddedModule", {
        "header": {
            "defaultValue": {
                "language": "en",
                "value": f"Wallet {wallet.public_pan}",
            }
        },
        "body": {
            "defaultValue": {
                "language": "en",
                "value": f"Balance: {money_filter(wallet.balance, wallet.currency)}",
            }
        },
        "uri": build_absolute_uri(order.event, "presale:event.order", {
            "order": order.pk, "secret": order.secret
        }),
    })]


def generate_apple_wallet_module(order_position: OrderPosition, order: Order):
    wallet = find_wallet(order_position, order)
    if not wallet:
        return []

    return [("backField", {
        "key": "wallet-number",
        "label": "Wallet number",
        "value": wallet.public_pan,
    }), ("backField", {
        "key": "wallet-balance",
        "label": "Wallet balance",
        "value": float(wallet.balance),
        "currencyCode": wallet.currency,
        "changeMessage": f"New balance for wallet {wallet.public_pan}: %@",
    })]


from django.core import validators
from django.core.exceptions import ValidationError
from pretix.base.models import LoggedModel, Device
from django.conf import settings
from django.db import models
import decimal
import random
import string

from pretix.base.settings import settings_hierarkey


def luhn_checksum(n: str):
    m = [0, 2, 4, 6, 8, 1, 3, 5, 7, 9]
    digits = list(n)
    odd_digits = [*map(int, digits[1:][::2])]
    even_digits = [m[int(d)] for d in digits[0:][::2]]
    x = sum(odd_digits) + sum(even_digits)

    x = (10 - x % 10)
    return 0 if x == 10 else x

def gen_wallet_pan(issuer):
    pan_len = issuer.settings.get("wallet_pan_length", 16, as_type=int)
    iin = issuer.settings.get("wallet_iin", "")
    while True:
        random_len = pan_len - 1 - len(iin)
        ian = "".join(random.choices(string.digits, k=random_len))
        pan = f"{iin}{ian}"
        pan = f"{pan}{luhn_checksum(pan)}"
        if not Wallet.objects.filter(pan=pan).exists():
            return pan


@settings_hierarkey.add(parent_field='issuer', cache_namespace='pretix_wallet')
class Wallet(LoggedModel):
    settings_namespace = 'pretix_wallet'

    issuer = models.ForeignKey(
        "pretixbase.Organizer",
        related_name="wallets",
        on_delete=models.PROTECT,
    )
    customer = models.OneToOneField(
        "pretixbase.Customer",
        related_name="wallet",
        null=True, blank=True,
        on_delete=models.SET_NULL
    )
    order_position = models.OneToOneField(
        "pretixbase.OrderPosition",
        related_name="wallet",
        null=True, blank=True,
        on_delete=models.SET_NULL
    )
    created_at = models.DateTimeField(auto_now_add=True)
    pan = models.CharField(max_length=19, validators=[validators.RegexValidator(regex=r"^[0-9]{10-19}$")], db_index=True)
    CURRENCY_CHOICES = [(c.alpha_3, c.alpha_3 + " - " + c.name) for c in settings.CURRENCIES]
    currency = models.CharField(max_length=10, choices=CURRENCY_CHOICES, validators=[
        validators.MinLengthValidator(3),
    ])

    def save(self, *args, **kwargs):
        if not self.pan:
            self.pan = gen_wallet_pan(self.issuer)

        super().save(*args, **kwargs)

    @property
    def public_pan(self):
        redacted_len = len(self.pan) - 8
        iin = self.pan[:4]
        last4 = self.pan[-4:]
        return f"{iin}{'*' * redacted_len}{last4}"

    @property
    def balance(self):
        if hasattr(self, 'cached_balance'):
            return self.cached_balance or decimal.Decimal('0.00')
        balance = self.transactions.aggregate(s=models.Sum('value'))['s'] or decimal.Decimal('0.00')
        self.cached_balance = balance
        return balance


class WalletTransaction(models.Model):
    wallet = models.ForeignKey(
        'Wallet',
        related_name='transactions',
        on_delete=models.PROTECT
    )
    timestamp = models.DateTimeField(auto_now_add=True)
    value = models.DecimalField(decimal_places=2, max_digits=13)
    order_position = models.ForeignKey(
        'pretixbase.OrderPosition',
        related_name="wallet_transactions",
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    order_payment = models.ForeignKey(
        'pretixbase.OrderPayment',
        related_name="wallet_transactions",
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    order_refund = models.ForeignKey(
        'pretixbase.OrderRefund',
        related_name="wallet_transactions",
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    descriptor = models.TextField(blank=True, null=False, default="")
    data = models.JSONField(default=dict)

    class Meta:
        ordering = ('-timestamp',)

    def device(self):
        if "_device_id" not in self.data:
            return None
        return Device.objects.get(pk=self.data["_device_id"])


class WalletItem(models.Model):
    item = models.OneToOneField('pretixbase.Item', on_delete=models.CASCADE, related_name="wallet")
    issue_wallet_balance = models.BooleanField(
        default=False, blank=True,
        help_text="Issue a wallet balance equivalent to this item's price"
    )

    def clean(self):
        if self.issue_wallet_balance:
            if self.item.admission:
                raise ValidationError("An item cannot both be an admission item and issue a balance")
            if self.item.issue_giftcard:
                raise ValidationError("An item cannot both issue a gift card and a balance")


from django.http import HttpResponse
import re

allowed_headers = (
    "accept",
    "authorization",
    "content-type",
    "user-agent",
    "x-csrftoken",
    "x-requested-with",
)

allowed_methods = (
    "DELETE",
    "GET",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
)

class CorsMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if response := self.preflight(request):
            self.add_response_headers(request, response)
            return response
        else:
            response = self.get_response(request)
            self.add_response_headers(request, response)
            return response

    def preflight(self, request):
        request._cors_enabled = self.is_enabled(request)
        if request._cors_enabled and request.method == "OPTIONS" and "access-control-request-method" in request.headers:
            return HttpResponse(headers={"content-length": "0"})
        else:
            return None

    def add_response_headers(self, request, response):
        enabled = getattr(request, "_cors_enabled", None)
        if enabled is None:
            enabled = self.is_enabled(request)

        if not enabled:
            return response

        response["Access-Control-Allow-Origin"] = "*"

        if request.method == "OPTIONS":
            response["Access-Control-Allow-Headers"] = ", ".join(allowed_headers)
            response["Access-Control-Allow-Methods"] = ", ".join(allowed_methods)

        return response

    @staticmethod
    def is_enabled(request):
        if request.path_info.startswith("/api/v1/device"):
            return True
        elif re.fullmatch(r"^/api/v1/organizers/([^/]+)/uic_keys/$", request.path_info):
            return True
        elif re.fullmatch(r"^/api/v1/organizers/([^/]+)/wallets/.*$", request.path_info):
            return True
        else:
            return False


from django.contrib import messages
from django.db import transaction
from django.db.models import OuterRef, Sum, Max, Subquery
from django.db.models.functions import Coalesce
from django.http import HttpResponseNotAllowed
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.functional import cached_property
from pretix.control.views.organizer import OrganizerDetailViewMixin, Organizer
from pretix.control.permissions import OrganizerPermissionRequiredMixin
from django.views.generic import ListView, FormView, DetailView
from . import models, forms, signals
import decimal


class SettingsView(OrganizerDetailViewMixin, OrganizerPermissionRequiredMixin, FormView):
    model = Organizer
    form_class = forms.WalletSettingsForm
    template_name = 'pretix_wallet/organizers/settings.html'
    permission = 'can_change_organizer_settings'

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['obj'] = self.request.organizer
        return kwargs

    def get_success_url(self):
        return reverse('plugins:pretix_wallet:settings', kwargs={
            'organizer': self.request.organizer.slug,
        })

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            form.save()
            if form.has_changed():
                self.request.organizer.log_action(
                    'pretix.organizer.settings', user=self.request.user, data={
                        k: form.cleaned_data.get(k) for k in form.changed_data
                    }
                )
                messages.success(self.request, "Your changes have been saved.")
            return redirect(self.get_success_url())
        else:
            messages.error(self.request, "We could not save your changes. See below for details.")
            return self.get(request)


class WalletListView(OrganizerDetailViewMixin, OrganizerPermissionRequiredMixin, ListView):
    model = models.Wallet
    template_name = 'pretix_wallet/organizers/wallets.html'
    permission = 'can_change_orders'
    context_object_name = 'wallets'
    paginate_by = 50

    def get_queryset(self):
        s = models.WalletTransaction.objects.filter(
            wallet=OuterRef('pk')
        ).order_by().values('wallet').annotate(s=Sum('value')).values('s')
        s_last_tx = models.WalletTransaction.objects.filter(
            wallet=OuterRef('pk')
        ).order_by().values('wallet').annotate(m=Max('timestamp')).values('m')
        qs = self.request.organizer.wallets.annotate(
            cached_balance=Coalesce(Subquery(s), decimal.Decimal('0.00')),
            last_tx=Subquery(s_last_tx),
        ).order_by('-created_at')
        if self.filter_form.is_valid():
            qs = self.filter_form.filter_qs(qs)
        return qs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['filter_form'] = self.filter_form
        return ctx

    @cached_property
    def filter_form(self):
        return forms.WalletFilterForm(data=self.request.GET, request=self.request)


class WalletView(OrganizerPermissionRequiredMixin, DetailView):
    model = models.Wallet
    template_name = 'pretix_wallet/organizers/wallet.html'
    permission = 'can_change_orders'
    context_object_name = 'wallet'
    slug_url_kwarg = "pan"
    slug_field = "pan"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['organizer'] = self.request.organizer
        ctx['charge_form'] = forms.WalletChargeForm()
        return ctx

    def get_queryset(self):
        return self.request.organizer.wallets.all()


class WalletSettingsView(OrganizerPermissionRequiredMixin, DetailView):
    model = models.Wallet
    template_name = 'pretix_wallet/organizers/wallet_settings.html'
    permission = 'can_change_organizer_settings'
    slug_url_kwarg = "pan"
    slug_field = "pan"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data()
        ctx['form'] = self.form
        ctx['organizer'] = self.request.organizer
        return ctx

    @cached_property
    def form(self):
        self.object = self.get_object()
        return forms.WalletIndividualSettingsForm(
            instance=self.object,
            data=self.request.POST if self.request.method == "POST" else None,
            customers=self.request.organizer.settings.customer_accounts and (
                self.request.user.has_organizer_permission(
                    self.request.organizer, 'can_manage_customers', request=self.request
                )
            ),
            initial={
                "wallet_minimum_balance": self.object.settings._cache().get("wallet_minimum_balance", None),
            }
        )

    def get_success_url(self):
        return reverse('plugins:pretix_wallet:wallet', kwargs={
            'organizer': self.request.organizer.slug,
            'pan': self.object.pan,
        })

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        if self.form.is_valid():
            self.form.save()

            if self.form.cleaned_data["wallet_minimum_balance"]:
                self.object.settings.set("wallet_minimum_balance", self.form.cleaned_data["wallet_minimum_balance"])
            else:
                self.object.settings.delete("wallet_minimum_balance")

            messages.success(self.request, "Your changes have been saved.")
            return redirect(self.get_success_url())
        else:
            messages.error(self.request, "We could not save your changes. See below for details.")
            return self.get(request)


class WalletManualChargeView(OrganizerPermissionRequiredMixin, DetailView):
    model = models.Wallet
    permission = 'can_change_orders'
    slug_url_kwarg = "pan"
    slug_field = "pan"

    def post(self, *args, **kwargs):
        wallet = self.get_object()
        form = forms.WalletChargeForm(self.request.POST)
        if form.is_valid():
            charged = False
            with transaction.atomic():
                if wallet.balance - form.cleaned_data["amount"] < wallet.settings.get("wallet_minimum_balance", as_type=decimal.Decimal):
                    messages.error(self.request, "Insufficient balance.")
                wallet.transactions.create(
                    value=-form.cleaned_data["amount"],
                    descriptor=form.cleaned_data["descriptor"] or "Charge",
                )
                charged = True

            if charged:
                signals.update_ticket_output.apply_async(kwargs={"wallet_pk": wallet.pk})
                messages.success(self.request, "The wallet has been charged.")
        else:
            messages.error(self.request, "The wallet could not be charged.")

        return redirect('plugins:pretix_wallet:wallet', organizer=self.request.organizer.slug, pan=wallet.pan)

    def get(self, *args, **kwargs):
        return HttpResponseNotAllowed(['POST'])



import decimal
from typing import Dict, Any, Union
from django.db import transaction
from django.http import HttpRequest
from django.template.loader import get_template
from pretix.base.models import Order, Customer, OrderPayment, OrderRefund
from pretix.base.payment import BasePaymentProvider, cart_session, PaymentException
from pretix.base.services.cart import add_payment_to_cart_session
from . import models, signals


class Wallet(BasePaymentProvider):
    identifier = "wallet"
    verbose_name = "Wallet"
    public_name = "Wallet balance"
    abort_pending_allowed = True
    multi_use_supported = True
    execute_payment_needs_user = False

    @staticmethod
    def customer(request, order: Order = None):
        if order and order.customer:
            return order.customer
        try:
            return request.organizer.customers.get(pk=cart_session(request).get('customer', -1))
        except Customer.DoesNotExist:
            return request.customer

    def is_allowed(self, request: HttpRequest, total: decimal.Decimal=None) -> bool:
        if not super().is_allowed(request, total):
            return False

        if customer := self.customer(request):
            if hasattr(customer, "wallet"):
                return True

        return False

    def payment_is_valid_session(self, request: HttpRequest) -> bool:
        if customer := self.customer(request):
            if hasattr(customer, "wallet"):
                return True

        return False

    def payment_form_render(self, request: HttpRequest, total: decimal.Decimal, order: Order=None) -> str:
        customer = self.customer(request, order)
        if not customer:
            return ""
        return get_template("pretix_wallet/payment/method_form.html").render({
            "request": request,
            "wallet": customer.wallet
        })

    def checkout_confirm_render(self, request, order: Order=None, info_data: dict=None) -> str:
        return get_template("pretix_wallet/payment/checkout_confirm.html").render({
            "request": request,
            "wallet": models.Wallet.objects.get(pk=info_data["wallet"])
        })

    def checkout_prepare(self, request: HttpRequest, cart: Dict[str, Any]) -> Union[bool, str]:
        wallet = self.customer(request).wallet
        cs = cart_session(request)
        add_payment_to_cart_session(
            cs,
            self,
            max_value=wallet.balance - wallet.settings.get("wallet_minimum_balance", as_type=decimal.Decimal),
            info_data={
                'wallet': wallet.pk
            }
        )
        return True

    def payment_prepare(self, request: HttpRequest, payment: OrderPayment) -> Union[bool, str]:
        customer = (payment.order.customer or self.customer(request))
        if not customer:
            return False
        wallet = customer.wallet
        payment.info_data = {
            'wallet': wallet.pk,
        }
        payment.amount = min(payment.amount, wallet.balance - wallet.settings.get("wallet_minimum_balance", as_type=decimal.Decimal))
        payment.save()
        return True

    def execute_payment(self, request: HttpRequest, payment: OrderPayment) -> str:
        wallet = models.Wallet.objects.get(pk=payment.info_data['wallet'])
        try:
            with transaction.atomic():
                if wallet.currency != self.event.currency:
                    raise PaymentException("The wallet is in the wrong currency")
                if wallet.balance - payment.amount < wallet.settings.get("wallet_minimum_balance", as_type=decimal.Decimal):
                    raise PaymentException("Insufficient balance.")

                trans = wallet.transactions.create(
                    value=-payment.amount,
                    descriptor=f"Payment for order #{payment.order.full_code}",
                    order_payment=payment,
                )
                signals.update_ticket_output.apply_async(kwargs={"wallet_pk": wallet.pk})
                payment.info_data = {
                    'wallet': wallet.pk,
                    'transaction_id': trans.pk,
                }
                payment.confirm()
        except PaymentException as e:
            payment.fail(info={'error': str(e)})
            raise e

    def api_payment_details(self, payment: OrderPayment):
        wallet = models.Wallet.objects.get(pk=payment.info_data['wallet'])
        return {
            'wallet': {
                'pan': wallet.pan,
                'public_pan': wallet.public_pan,
            }
        }

    def api_refund_details(self, refund: OrderRefund):
        wallet = models.Wallet.objects.get(pk=refund.info_data['wallet'])
        return {
            'wallet': {
                'pan': wallet.pan,
                'public_pan': wallet.public_pan,
            }
        }

    def payment_control_render(self, request, payment: OrderPayment) -> str:
        wallet = models.Wallet.objects.get(pk=payment.info_data['wallet'])
        return get_template("pretix_wallet/payment/payment_control.html").render({
            "request": request,
            "wallet": wallet
        })

    def refund_control_render(self, request, refund: OrderRefund) -> str:
        wallet = models.Wallet.objects.get(pk=refund.info_data['wallet'])
        return get_template("pretix_wallet/payment/payment_control.html").render({
            "request": request,
            "wallet": wallet
        })

    def payment_control_render_short(self, payment: OrderPayment) -> str:
        wallet = models.Wallet.objects.get(pk=payment.info_data['wallet'])
        return wallet.pan

    def refund_control_render_short(self, refund: OrderRefund) -> str:
        wallet = models.Wallet.objects.get(pk=refund.info_data['wallet'])
        return wallet.pan

    def payment_partial_refund_supported(self, payment: OrderPayment) -> bool:
        return True

    def payment_refund_supported(self, payment: OrderPayment) -> bool:
        return True
    
    def execute_refund(self, refund: OrderRefund):
        with transaction.atomic():
            wallet = models.Wallet.objects.get(pk=refund.info_data.get('wallet') or refund.payment.info_data.get('wallet'))
            trans = wallet.transactions.create(
                value=refund.amount,
                descriptor=f"Refund for order #{refund.order.full_code}",
                order_refund=refund,
            )
            signals.update_ticket_output.apply_async(kwargs={"wallet_pk": wallet.pk})
            refund.info_data = {
                'wallet': wallet.pk,
                'transaction_id': trans.pk,
            }
            refund.done()


from django.urls import path
from pretix.api import urls
from . import views, api


urlpatterns = [
    path("control/organizer/<organizer>/wallets/", views.WalletListView.as_view(), name='wallets'),
    path("control/organizer/<organizer>/wallets/wallet/<str:pan>/", views.WalletView.as_view(), name='wallet'),
    path("control/organizer/<organizer>/wallets/wallet/<str:pan>/settings/", views.WalletSettingsView.as_view(), name='wallet_settings'),
    path("control/organizer/<organizer>/wallets/wallet/<str:pan>/manual_charge/", views.WalletManualChargeView.as_view(), name='wallet_manual_charge'),
    path("control/organizer/<organizer>/wallets/settings/", views.SettingsView.as_view(), name='settings'),
]

urls.orga_router.register('wallets', api.WalletViewSet, basename='wallets')


from django.utils.translation import gettext_lazy

from . import __version__

try:
    from pretix.base.plugins import PluginConfig, PLUGIN_LEVEL_ORGANIZER
except ImportError:
    raise RuntimeError("Please use pretix 2.7 or above to run this plugin!")


class PluginApp(PluginConfig):
    default = True
    name = "pretix_wallet"
    verbose_name = "Wallet"

    class PretixPluginMeta:
        name = gettext_lazy("Wallet")
        author = "AS207960 Cyfyngedig"
        description = gettext_lazy("Allows users to pre-buy wallet balances at order time e.g. for an event bar.")
        visible = True
        experimental = True
        version = __version__
        category = "FEATURE"
        compatibility = "pretix>=2.7.0"
        # level = PLUGIN_LEVEL_ORGANIZER
        settings_links = []
        navigation_links = []

    def ready(self):
        from django.conf import settings
        settings.MIDDLEWARE.append("pretix_wallet.middleware.CorsMiddleware")

        from . import signals  # NOQA



import decimal

from django.db import transaction
from django.dispatch import receiver
from django.template.loader import get_template
from django.urls import resolve, reverse
from pretix.base.models.orders import Order
from pretix.base.services import tickets
from pretix.base.signals import register_payment_providers, customer_created, order_paid, order_changed
from pretix.control.signals import nav_organizer, item_forms
from pretix.presale.signals import order_info_top
from pretix.celery_app import app
from pretix_uic_barcode.signals import register_barcode_element_generators, register_vas_element_generators, generate_google_wallet_module, generate_apple_wallet_module
from pretix_uic_barcode import ticket_output

from . import payment, models, elements, forms


@receiver(register_payment_providers, dispatch_uid="payment_wallet")
def register_payment_provider(sender, **kwargs):
    return payment.Wallet


@receiver(nav_organizer, dispatch_uid="wallet_organav")
def control_nav_orga_import(sender, request=None, **kwargs):
    url = resolve(request.path_info)
    if not request.user.has_organizer_permission(request.organizer, "can_change_orders", request=request):
        return []
    if not request.organizer.events.filter(plugins__icontains="pretix_wallet"):
        return []
    return [
        {
            'label': "Wallets",
            'url': reverse('plugins:pretix_wallet:wallets', kwargs={
                'organizer': request.organizer.slug,
            }),
            'icon': 'university',
            'children': [{
                'label': "Wallets",
                'url': reverse('plugins:pretix_wallet:wallets', kwargs={
                    'organizer': request.organizer.slug
                }),
                'active': url.url_name == "wallets",
            }, {
                'label': "Settings",
                'url': reverse('plugins:pretix_wallet:settings', kwargs={
                    'organizer': request.organizer.slug
                }),
                'active': url.url_name == "settings",
            }]
        }
    ]


@receiver(customer_created, dispatch_uid="wallet_customer_created")
def customer_created(sender, customer, **kwargs):
    if sender.settings.wallet_create_for_customers and not hasattr(customer, "wallet"):
        models.Wallet.objects.create(
            issuer=sender,
            customer=customer,
            currency=sender.settings.wallet_default_currency,
        )


@receiver(order_info_top, dispatch_uid="wallet_show_balance")
def order_info_balance(sender, request, order, **kwargs):
    template = get_template("pretix_wallet/order/balance.html")

    wallets = []
    if hasattr(order.customer, "wallet"):
        wallets.append(order.customer.wallet)

    ctx = {
        'order': order,
        'request': request,
        'event': sender,
        'wallets': wallets,
    }
    return template.render(ctx, request=request)


@receiver(item_forms, dispatch_uid="wallet_item_issue_balance")
def item_issue_balance(sender, item, request, **kwargs):
    try:
        inst = models.WalletItem.objects.get(item=item)
    except  models.WalletItem.DoesNotExist:
        inst = models.WalletItem(item=item)
    return forms.WalletItemForm(
        instance=inst,
        data=(request.POST if request.method == "POST" else None),
        prefix="wallet"
    )


@receiver(order_paid, dispatch_uid="wallet_order_paid_issue_balance")
@receiver(order_changed, dispatch_uid="wallet_order_changed_issue_balance")
@transaction.atomic()
def order_issue_balance(sender, order, **kwargs):
    if order.status != Order.STATUS_PAID:
        return
    any_wallets = False
    for p in order.positions.all():
        if hasattr(p.item, "wallet") and p.item.wallet.issue_wallet_balance:
            issued = decimal.Decimal('0.00')
            for wt in p.wallet_transactions.all().distinct():
                issued += wt.value
            tbi = p.price - issued
            if tbi > 0:
                any_wallets = True

                if order.customer:
                    if hasattr(order.customer, "wallet"):
                        wallet = order.customer.wallet
                    else:
                        wallet = models.Wallet.objects.create(
                            issuer=sender.organizer,
                            customer=order.customer,
                            currency=sender.settings.wallet_default_currency,
                        )
                        wallet.save()
                else:
                    if hasattr(order, "wallet"):
                        wallet = order.wallet
                    else:
                        wallet = models.Wallet.objects.create(
                            issuer=sender.organizer,
                            order_position=p,
                            currency=sender.settings.wallet_default_currency,
                        )
                        wallet.save()

                wallet.transactions.create(value=tbi, order_position=p, descriptor=f"Order #{order.full_code}")
                update_ticket_output.apply_async(kwargs={"wallet_pk": wallet.pk})

    if any_wallets:
        tickets.invalidate_cache.apply_async(kwargs={'event': sender.pk, 'order': order.pk})


@receiver(register_barcode_element_generators, dispatch_uid="wallet_barcode_element_generator")
def barcode_element_generator(sender, **kwargs):
    return [elements.WalletBarcodeElementGenerator]


@receiver(register_vas_element_generators, dispatch_uid="wallet_vas_element_generator")
def vas_element_generator(sender, **kwargs):
    return [elements.WalletVASElementGenerator]


@receiver(generate_google_wallet_module, dispatch_uid="wallet_google_module_generator")
def google_module_generator(sender, **kwargs):
    return [elements.generate_google_wallet_module]


@receiver(generate_apple_wallet_module, dispatch_uid="wallet_apple_module_generator")
def apple_module_generator(sender, **kwargs):
    return [elements.generate_apple_wallet_module]


@app.task(acks_late=True)
def update_ticket_output(wallet_pk):
    wallet = models.Wallet.objects.get(pk=wallet_pk)
    if wallet.order_position:
        ticket_output.update_ticket_output.apply_async(kwargs={"event": wallet.order_position.event.pk, "position_pk": wallet.order_position.pk})
    
    if wallet.customer:
        for order in wallet.customer.orders.all():
            ticket_output.update_ticket_output_all.apply_async(kwargs={"event": order.event.pk, "order_pk": order.pk})
        


import decimal
from django.db import transaction
from django.http import Http404
from i18nfield.rest_framework import I18nAwareModelSerializer
from pretix.base.models import Device, TeamAPIToken
from pretix.helpers import OF_SELF
from rest_framework import viewsets, serializers, status
from rest_framework.decorators import action
from rest_framework.exceptions import MethodNotAllowed
from rest_framework.response import Response
from . import models, signals


class WalletSerializer(I18nAwareModelSerializer):
    balance = serializers.DecimalField(max_digits=13, decimal_places=2, read_only=True)
    issuer = serializers.SlugRelatedField(slug_field='slug', read_only=True)
    customer = serializers.SlugRelatedField(slug_field='identifier', read_only=True)

    class Meta:
        model = models.Wallet
        fields = ('id', 'issuer', 'balance', 'customer', 'created_at', 'pan',
                  'public_pan', 'currency')


class WalletViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = WalletSerializer
    queryset = models.Wallet.objects.none()
    permission = 'can_change_orders'

    def get_queryset(self):
        return self.request.organizer.wallets.all()

    def get_object(self):
        try:
            return self.get_queryset().get(pan=self.kwargs['pk'])
        except models.Wallet.DoesNotExist:
            raise Http404

    def get_serializer_context(self):
        ctx = super().get_serializer_context()
        ctx['organizer'] = self.request.organizer
        return ctx

    def perform_destroy(self, instance):
        raise MethodNotAllowed("Wallets cannot be deleted.")

    @action(detail=True, methods=["POST"])
    @transaction.atomic
    def charge(self, request, **kwarg):
        wallet = models.Wallet.objects.select_for_update(of=OF_SELF).get(pk=self.get_object().pk)
        amount = serializers.DecimalField(max_digits=13, decimal_places=2).to_internal_value(request.data.get('amount'))
        descriptor = serializers.CharField(allow_blank=True, allow_null=True).to_internal_value(request.data.get('descriptor', ''))
        data = serializers.JSONField(required=False, allow_null=True).to_internal_value(request.data.get('data', {}))
        if isinstance(request.auth, Device):
            data["_device_id"] = request.auth.pk
        elif isinstance(request.auth, TeamAPIToken):
            data["_team_token_id"] = request.auth.pk
        else:
            data["_user_id"] = request.user.pk
        if wallet.balance - amount < wallet.settings.get("wallet_minimum_balance", as_type=decimal.Decimal):
            return Response({
                "amount": ["Insufficient balance"],
            }, status=status.HTTP_402_PAYMENT_REQUIRED)
        wallet.transactions.create(
            value=-amount,
            descriptor=descriptor or "Charge",
            data=data,
        )
        signals.update_ticket_output.apply_async(kwargs={"wallet_pk": wallet.pk})
        return Response(WalletSerializer(self.get_object(), context=self.get_serializer_context()).data, status=status.HTTP_200_OK)


# Generated by Django 4.2.23 on 2025-08-07 14:45

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('pretix_wallet', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wallet',
            name='pan',
            field=models.CharField(db_index=True, max_length=19),
        ),
        migrations.AlterField(
            model_name='wallettransaction',
            name='order_payment',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='wallet_transactions', to='pretixbase.orderpayment'),
        ),
        migrations.AlterField(
            model_name='wallettransaction',
            name='order_position',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='wallet_transactions', to='pretixbase.orderposition'),
        ),
        migrations.AlterField(
            model_name='wallettransaction',
            name='order_refund',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='wallet_transactions', to='pretixbase.orderrefund'),
        ),
    ]



from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator
from django.db.models import F, Q
from django.urls import reverse
from django_scopes.forms import SafeModelChoiceField
from pretix.base.forms import SettingsForm
from pretix.control.forms.filter import FilterForm
from django import forms
from pretix.control.forms.widgets import Select2 as BaseSelect2

from . import models


class WalletItemForm(forms.ModelForm):
    title = "Wallet"

    class Meta:
        model = models.WalletItem
        fields = ('issue_wallet_balance',)


class WalletSettingsForm(SettingsForm):
    wallet_create_for_customers = forms.BooleanField(
        label="Create wallet on customer account creation - i.e. before an order is placed",
        required=False
    )
    wallet_default_currency = forms.ChoiceField(
        label="Default currency",
        choices=[(c.alpha_3, c.alpha_3 + " - " + c.name) for c in settings.CURRENCIES],
    )
    wallet_iin = forms.CharField(
        max_length=18,
        label="Issuer Identification Number"
    )
    wallet_pan_length = forms.IntegerField(
        label="PAN length",
        min_value=10,
        max_value=19,
        initial=16,
    )
    wallet_minimum_balance = forms.DecimalField(
        label="Default minimum balance",
        decimal_places=2,
        max_digits=13,
        initial=0,
        validators=[MaxValueValidator(0)],
    )

    def clean(self):
        if self.cleaned_data.get("wallet_create_for_customers") and not self.cleaned_data.get("wallet_default_currency"):
            raise ValidationError({
                "wallet_default_currency": "Default currency must be specified",
            })

        if self.cleaned_data.get("wallet_pan_length", 16) - len(self.cleaned_data.get("wallet_iin", "")) < 2:
            raise ValidationError({
                "wallet_iin": "IIN is too long for the chosen PAN length",
            })


class Select2(BaseSelect2):
    def build_attrs(self, base_attrs, extra_attrs=None):
        attrs = super().build_attrs(base_attrs, extra_attrs)
        if self.choices.field.empty_label:
            attrs["data-placeholder"] = self.choices.field.empty_label
        return attrs


class WalletIndividualSettingsForm(forms.ModelForm):
    wallet_minimum_balance = forms.DecimalField(
        label="Minimum balance",
        decimal_places=2,
        max_digits=13,
        initial=None,
        validators=[MaxValueValidator(0)],
        required=False,
        help_text="Set an empty value to use the organizer default",
    )

    class Meta:
        model = models.Wallet
        fields = ('customer',)
        field_classes = {
            'customer': SafeModelChoiceField,
        }

    def __init__(self, *args, **kwargs):
        customers = kwargs.pop('customers')
        super().__init__(*args, **kwargs)

        if customers:
            self.fields['customer'].queryset = self.instance.issuer.customers.all()
            self.fields['customer'].widget = Select2(
                attrs={
                    'data-model-select2': 'generic',
                    'data-select2-url': reverse('control:organizer.customers.select2', kwargs={
                        'organizer': self.instance.issuer.slug,
                    }),
                }
            )
            self.fields['customer'].widget.choices = self.fields['customer'].choices
            self.fields['customer'].required = False
        else:
            del self.fields['customer']


class WalletChargeForm(forms.Form):
    descriptor = forms.CharField(required=False)
    amount = forms.DecimalField(
        max_digits=13, decimal_places=2,
        help_text="Enter a negative amount for top-ups",
    )


class WalletFilterForm(FilterForm):
    orders = {
        'created_at': 'created_at',
        'customer': 'customer__name_cached',
        'last_tx': F('last_tx').asc(nulls_first=True),
        '-last_tx': F('last_tx').desc(nulls_last=True),
        'id': 'pan',
        'value': 'cached_value',
    }
    state = forms.ChoiceField(
        label="Status",
        choices=(
            ('', "All"),
            ('empty', "Empty"),
            ('with_value', "With value"),
        ),
        required=False
    )
    query = forms.CharField(
        label="Search query",
        widget=forms.TextInput(attrs={
            'placeholder': "Query"
        }),
        required=False
    )

    def __init__(self, *args, **kwargs):
        kwargs.pop('request')
        super().__init__(*args, **kwargs)

    def filter_qs(self, qs):
        fdata = self.cleaned_data

        if fdata.get('query'):
            query = fdata.get('query')

            qs = qs.filter(
                Q(pan__icontains=query)
                | Q(transactions__descriptor__icontains=query)
                | Q(customer__name_cached__icontains=query)
                | Q(transactions__order_payment__order__code__icontains=query)
                | Q(transactions__order_refund__order__code__icontains=query)
            )
        if fdata.get('state') == 'empty':
            qs = qs.filter(cached_balance=0)
        elif fdata.get('state') == 'with_value':
            qs = qs.exclude(cached_balance=0)

        if fdata.get('ordering'):
            qs = qs.order_by(self.get_order_by())
        else:
            qs = qs.order_by('-created_at')

        return qs.distinct()